{
  "type": "record",
  "name": "CorpUserKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a CorpUser",
  "fields": [
    {
      "name": "username",
      "type": "string",
      "doc": "The name of the AD/LDAP user.",
      "Searchable": {
        "boostScore": 2.0,
        "enableAutocomplete": true,
        "fieldName": "ldap",
        "fieldType": "TEXT_PARTIAL"
      }
    }
  ],
  "Aspect": {
    "name": "corpUserKey"
  }
}