{
  "type": "record",
  "name": "DataFlowKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a Data Flow",
  "fields": [
    {
      "name": "orchestrator",
      "type": "string",
      "doc": "Workflow manager like azkaban, airflow which orchestrates the flow",
      "Searchable": {
        "fieldType": "TEXT_PARTIAL"
      }
    },
    {
      "name": "flowId",
      "type": "string",
      "doc": "Unique Identifier of the data flow",
      "Searchable": {
        "enableAutocomplete": true,
        "fieldType": "TEXT_PARTIAL"
      }
    },
    {
      "name": "cluster",
      "type": "string",
      "doc": "Cluster where the flow is executed",
      "Searchable": {
        "fieldType": "TEXT_PARTIAL"
      }
    }
  ],
  "Aspect": {
    "name": "dataFlowKey"
  }
}